# -*- coding: utf-8 -*-

#
# repartidor.py - reparte trabajos en distintos hilos
#
# Copyright (C) 2009 - Facundo Batista <facundo@taniquetil.com.ar>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import threading, Queue, time

class Trabajador(threading.Thread):
    '''Clase que usa el repartidor internamente.

    Ejecuta una funcin recibida de afuera.
    '''
    def __init__(self, nro, funcion, colaInput, colaOutput, termine):
        self.yo = nro
        self.funcion = funcion
        self.cinp = colaInput
        self.cout = colaOutput
        self.termine = termine
        threading.Thread.__init__(self)

    def run(self):
        while True:
            info = self.cinp.get()
            if info == "quit":
                break
            res = self.funcion(*info)
            self.termine.set()
            self.cout.put((info, res))


class Repartidor:
    '''Clase que reparte muchas tareas en algunos hilos.

    Recibe la funcion a ejecutar, la cantidad de trabajadores
    a abrir, y opcionalmente una cola Queue en la que ira
    dejando mensajes de lo que va haciendo.
    '''
    def __init__(self, funcion, cant, logq=None):
        self._cantw = cant
        self.logq = logq

        # lanzamos los n hilos para cada destino
        self.qEnviar = [Queue.Queue() for x in range(self._cantw)]
        self.qRecbir = [Queue.Queue() for x in range(self._cantw)]
        self.eTermin = [threading.Event() for x in range(self._cantw)]
        for i in range(self._cantw):
            h = Trabajador(i, funcion, self.qEnviar[i], self.qRecbir[i], self.eTermin[i])
            h.start()
        if self.logq is not None:
            self.logq.put("repartidor: %s - Se crearon %d hilos" % (time.ctime(), cant))

    def procesar(self, trabajos, funcAviso=None):
        """ Procesa los trabajos recibidos.

        Los desparrama entre los trabajadores, en paralelo, mientras estén libres. Va
        juntando los resultados y va entregando los resultados

        El primer argumento son todos los trabajos, y el segundo una funcion que
        se ejecuta cada vez que termina un trabajo (opcional)
        """

        # preparamos
        encolados = trabajos[:]
        encolados.reverse()
        disponibles = [True] * self._cantw

        # ejecutamos mientras tengamos encolados pendientes o haya un destino sin terminar
        while encolados or sum(disponibles) < self._cantw:
            if self.logq is not None:
                self.logq.put("repartidor: %s - Hay encolados (%d) o estamos esperando algun trabajo (%r)" % (time.ctime(), len(encolados), disponibles))

            # si hay algún hilo libre le damos trabajo (si hay)
            while (encolados and (True in disponibles)):
                payload = encolados.pop()
                libre = disponibles.index(True)
                q = self.qEnviar[libre]
                q.put(payload)
                disponibles[libre] = False
                if self.logq is not None:
                    self.logq.put("repartidor: %s - Enviamos %r al hilo %d" % (time.ctime(), payload, libre))

            # revisamos los pendientes, para ver si terminó alguno
            for i in range(self._cantw):
                if not self.eTermin[i].isSet():
                    continue

                # tenemos un dato de alguno
                result = self.qRecbir[i].get()
                self.eTermin[i].clear()
                if self.logq is not None:
                    self.logq.put("repartidor: %s - Recibimos %r del hilo %d" % (time.ctime(), result, i))
                yield result
                if funcAviso is not None:
                    funcAviso()
                disponibles[i] = True

            # dormimos para que el while no me ocupe todo el procesador
            time.sleep(.3)



    def fin(self):
        '''Cierra todos los threads.'''
        # terminamos todo, limpiamos y nos vamos
        for q in self.qEnviar:
            q.put("quit")
        if self.logq is not None:
            self.logq.put("repartidor: %s - Se envio quit a todos los hilos" % time.ctime())
        return


__all__ = [Repartidor]
